"""
Clients for the dataporten APIs
"""
import logging
import urllib

import requests
import requests.auth


logger = logging.getLogger(__name__)


def quote_path_arg(arg):
    return urllib.parse.quote_plus(str(arg))


def urljoin(base_url, *paths):
    """An urljoin that takes multiple path arguments.

    Note how urllib.parse.urljoin will assume 'relative to parent' when the
    base_url doesn't end with a '/':

    >>> urllib.parse.urljoin('https://localhost/foo', 'bar')
    'https://localhost/bar'

    >>> urljoin('https://localhost/foo', 'bar')
    'https://localhost/foo/bar'

    >>> urljoin('https://localhost/foo', 'bar', 'baz')
    'https://localhost/foo/bar/baz'
    """
    for path in paths:
        base_url = urllib.parse.urljoin(base_url.rstrip("/") + "/", path)
    return base_url


def merge_dicts(*dicts):
    """
    Combine a series of dicts.

    >>> merge_dicts({'a': 1}, {'b': 2})
    {'a': 1, 'b': 2}
    >>> merge_dicts({'a': 1}, {'a': 2})
    {'a': 2}

    None-values are ignored:

    >>> merge_dicts(None, None, None)
    {}

    If a dict-value is None, the key is deleted:

    >>> merge_dicts({'a': 1, 'b': 2, 'c': 3}, {'b': None})
    {'a': 1, 'c': 3}
    """
    combined = dict()
    for d in dicts:
        if not d:
            continue
        for k in d:
            if d[k] is None:
                if combined[k]:
                    del combined[k]
                else:
                    continue
            else:
                combined[k] = d[k]
    return combined


class HttpBearerAuth(requests.auth.AuthBase):
    """
    Bearer authorization for requests.

    Example:

    >>> auth = HttpBearerAuth('e64a7...')
    >>> str(auth)
    'Bearer e64a7...'
    >>> r = requests.Request()
    >>> auth(r).headers
    {'Authorization': 'Bearer e64a7...'}

    Use with requests:

    ::

        requests.get('https://example.org/', auth=HttpBearerAuth('e64a7...'))

    """

    header = "Authorization"
    scheme = "Bearer"

    def __init__(self, token, header=header, scheme=scheme):
        self.header = str(header)
        self.scheme = str(scheme)
        self.token = str(token)

    def __repr__(self):
        return ("<{cls.__name__} " "'{scheme} ...' at 0x{addr:02x}>").format(
            cls=type(self), scheme=str(self.scheme), addr=id(self)
        )

    def __str__(self):
        return self.scheme + " " + self.token

    def __eq__(self, other):
        attrs = ("token", "scheme", "header")
        if not all(hasattr(other, a) for a in attrs):
            return NotImplemented
        return all(getattr(self, a) == getattr(other, a) for a in attrs)

    def __ne__(self, other):
        eq_result = self == other
        if eq_result is NotImplemented:
            return NotImplemented
        else:
            return not eq_result

    def __call__(self, r):
        r.headers[self.header] = str(self)
        return r


class DataportenEndpoints(object):
    """Get and format urls to the Feide/Dataporten APIs."""

    auth_base = "https://auth.dataporten.no"
    api_base = "https://api.dataporten.no"

    def __init__(self, auth_url=None, api_url=None):
        self.auth_base = auth_url or self.auth_base
        self.api_base = api_url or self.api_base

    @property
    def userinfo(self):
        return urljoin(self.auth_base, "openid/userinfo")

    @property
    def extended_userinfo(self):
        return urljoin(self.api_base, "userinfo/v1/userinfo")

    @property
    def logout(self):
        return urljoin(self.auth_base, "logout")

    def get_photo_url(self, photo_id):
        photo_id = quote_path_arg(photo_id)
        return urljoin(self.api_base, "userinfo/v1/user/media", photo_id)


class DataportenClient(object):

    """
    A client to communicate with the Feide APIs.

    To use this client, an access token authorized for the relevant scopes must
    be provided.
    """

    default_headers = {
        "Accept": "application/json",
    }

    def __init__(self, token, headers=None):
        self.url = DataportenEndpoints()
        self.auth = HttpBearerAuth(token)
        self.headers = merge_dicts(self.default_headers, headers)

    @property
    def token(self):
        return self.auth.token

    @token.setter
    def token(self, value):
        self.auth.token = value

    def _get(self, url, headers=None):
        headers = merge_dicts(self.headers, headers)
        logger.debug("Calling GET %s", url)
        return requests.get(url, auth=self.auth, headers=headers)

    def get_user_info(self):
        """
        get user info.

        The response varies depending on which scopes the access token is
        authorized for.
        """
        return self._get(self.url.userinfo)

    def get_photo(self, photo_id):
        """
        get user photo

        Requires the Feide ``profile`` scope.
        """
        return self._get(self.url.get_photo_url(photo_id))

    def get_extended_user_info(self):
        """
        Get extended user info.

        The response varies depending on which scopes the access token is
        authorized for.
        """
        return self._get(self.url.extended_userinfo)


class DataportenApi(object):
    # TODO:
    # Should we verify perform a userinfo lookup using token and compare
    # client_id/user_id on auth?

    def __init__(self, user, app=None):
        self.user = user
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        pass

    @property
    def client(self):
        return DataportenClient(self.user.access_token)

    def get_user_info(self):
        result = self.client.get_user_info()
        result.raise_for_status()
        data = result.json()
        if data["audience"] != self.user.client_id:
            raise RuntimeError("Invalid audience from userinfo")
        return data

    def get_user_photo(self):
        user_info = self.get_user_info()
        if "user" not in user_info:
            raise ValueError("missing user info from the response")
        user_data = user_info["user"]
        if "profilephoto" not in user_data:
            raise ValueError("missing user profilephoto in user data")
        photo_id = user_data["profilephoto"]
        result = self.client.get_photo(photo_id)
        result.raise_for_status()
        return result

    def get_extended_user_info(self):
        """Get the extended user info."""
        result = self.client.get_extended_user_info()
        result.raise_for_status()
        return result.json()
