"""
Utilities to fix issues from using proxies in front of a dataporten API.

This module provides :py:class:`ProxyFix`, which modifies
:py:data:`flask.request.remote_addr` according to the ``X-Forwarded-For``
header from *trusted* proxies.

To skip all loopback addresses when setting ``REMOTE_ADDR`` from
``X-Forwarded-For``, you would need to:

::

    app = Flask('example')
    loopback_fix = ProxyFix(['127.0.0.0/8', '::1'])
    loopback_fix.init_app(app)

Note that ``REMOTE_ADDR`` is only modified if the request also comes from a
whitelisted address.

"""
import ipaddress
import logging


logger = logging.getLogger(__name__)

environ_backup_prefix = __name__ + '.orig_'


# # TODO: Maybe -- in stead of applying wsgi middleware to alter the
# #       REMOTE_ADDR envvar, this should be a flask extension that simply
# #       presents an alternative remote_addr, so that we could do:
#
# proxy_loopback = ProxyView(['127.0.0.0/8', '::1'])
# proxy_gateway = ProxyView([<gatekeeper-subnets...>], before=proxy_loopback)
#
# # and then ask:
#
# proxy_loopback.remote_addr  # client ip seen from the app server/nginx.
#                             # probably a gatekeeper address
# proxy_gateway.remote_addr   # client ip seen from the gatekeeper


class ProxyFix(object):
    """
    WSGI middleware to update ``REMOTE_ADDR`` according to ``X-Forwarded-For``.

    This middleware application is based on
    :py:class:`werkzeug.contrib.fixers.ProxyFix`. Rather than fetching the
    ``-n``-th item from ``X-Forwarded-For``, it removes white-listed proxy IPs,
    if the request comes from one of those proxies.
    """

    def __init__(self, whitelist, app=None):
        self.whitelist = tuple(ipaddress.ip_network(subnet)
                               for subnet in (whitelist or ()))
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.wrapped = app.wsgi_app
        app.wsgi_app = self

    def is_whitelisted(self, ip_addr):
        """ Check if an ip address is whitelisted. """
        ip_addr = ipaddress.ip_address(ip_addr)
        for subnet in self.whitelist:
            if ip_addr in subnet:
                logger.debug('%r is whitelisted (by %r)', ip_addr, subnet)
                return True
        return False

    def get_remote_addr(self, forwarded_for):
        """ Get ``REMOTE_ADDR`` given a list of ``X-Forwarded-For`` IPs. """
        for ip_addr in reversed(forwarded_for):
            if self.is_whitelisted(ip_addr):
                continue
            return ip_addr

    def update_environ(self, environ):
        """ Update wsgi environment according to ``X-Forwarded-For``. """
        getter = environ.get

        def update_value(key, value):
            orig_key = environ_backup_prefix + key.lower()
            orig_val = getter(key)
            environ[orig_key] = environ[key]
            if value:
                logger.debug('environ[%r], %r -> %r', key, orig_val, value)
                environ[key] = value

        forwarded_proto = getter('HTTP_X_FORWARDED_PROTO', '')
        forwarded_for = tuple(
            x.strip()
            for x in getter('HTTP_X_FORWARDED_FOR', '').split(',')
            if x.strip())
        forwarded_host = getter('HTTP_X_FORWARDED_HOST', '')

        # forwarded_for = [x for x in [x.strip() for x in forwarded_for] if x]
        remote_addr = self.get_remote_addr(forwarded_for)

        update_value('REMOTE_ADDR', remote_addr)
        update_value('HTTP_HOST', forwarded_host)
        update_value('wsgi.url_scheme', forwarded_proto)

    def __call__(self, environ, start_response):
        """ wsgi wrapper app.

        Updates the environment with a new ``REMOTE_ADDR`` if the current
        ``REMOTE_ADDR`` is whitelisted.
        """
        if self.is_whitelisted(environ['REMOTE_ADDR']):
            self.update_environ(environ)
        return self.wrapped(environ, start_response)
